/*************************************************************************/
/*                                                                       */
/*                Centre for Speech Technology Research                  */
/*                     University of Edinburgh, UK                       */
/*                         Copyright (c) 1996                            */
/*                        All Rights Reserved.                           */
/*                                                                       */
/*  Permission to use, copy, modify, distribute this software and its    */
/*  documentation for research, educational and individual use only, is  */
/*  hereby granted without fee, subject to the following conditions:     */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*  This software may not be used for commercial purposes without        */
/*  specific prior written permission from the authors.                  */
/*                                                                       */
/*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/
/*             Author :  Paul Taylor and Alan Black                      */
/*             Date   :  May 1996                                        */
/*-----------------------------------------------------------------------*/
/*            EST_Wave Class source file                                 */
/*                                                                       */
/*=======================================================================*/

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "EST_cutils.h"
#include "EST_Wave.h"
#include "EST_wave_utils.h"
#include "EST_TNamedEnum.h"
#include "EST_WaveFile.h"

#include "waveP.h"

const EST_String DEF_FILE_TYPE = "nist";
const EST_String DEF_SAMPLE_TYPE = "short";

const int EST_Wave::default_sample_rate=16000;

static EST_TValuedEnumDefinition<EST_sample_type_t, const char *,NO_INFO> st_names[] = {
{st_unknown,  {"undef"}},
{st_schar, {"schar"}},
{st_uchar, {"uchar"}},
{st_short, {"short"}},
{st_shorten, {"shorten"}},
{st_int, {"int"}},
{st_float, {"float"}},
{st_double, {"double"}},
{st_mulaw, {"mulaw"}},
{st_adpcm, {"adpcm"}},
{st_alaw, {"alaw"}},
{st_ascii, {"ascii"}},
{st_unknown, {0}}
};

static EST_TNamedEnum<EST_sample_type_t> sample_type_map(st_names);

#if defined(INSTANTIATE_TEMPLATES)
#include "../base_class/EST_TNamedEnum.cc"
template class EST_TNamedEnum<EST_sample_type_t>; 
template class EST_TNamedEnumI<EST_sample_type_t, NO_INFO>; 
template class EST_TValuedEnum<EST_sample_type_t,const char *>; 
template class EST_TValuedEnumI<EST_sample_type_t,const char *, NO_INFO>; 
#endif

EST_Wave::EST_Wave()
{
    default_vals();
}

EST_Wave::EST_Wave(const EST_Wave &w)
{
    default_vals();
    copy(w);
}

EST_Wave::EST_Wave(int n, int c, int sr)
{
  default_vals(n,c);
  set_sample_rate(sr);
}

EST_Wave::EST_Wave(int samps, int chans,
		   short *memory, int offset, int sample_rate, 
		   int free_when_destroyed)
{
  default_vals();
  p_values.set_memory(memory, offset, samps, chans, free_when_destroyed);
  set_sample_rate(sample_rate);
}

void EST_Wave::default_vals(int n, int c) 
{
    // real defaults
  p_values.resize(n,c);

  p_sample_rate = default_sample_rate;
  p_file_type = "nist";
  p_name = "";
}

void EST_Wave::free_wave()
{
    p_values.resize(0,0);
}

EST_Wave::~EST_Wave()
{
}

void EST_Wave::copy_setup(const EST_Wave &w)
{
    p_sample_rate = w.p_sample_rate;
    p_name = w.p_name;
}

void EST_Wave::copy_data(const EST_Wave &w)
{
    p_values.copy(w.p_values);
}

void EST_Wave::copy(const EST_Wave &w)
{
    copy_setup(w);

    copy_data(w);
}

short &EST_Wave::a(int i, int channel)
{
  if (i<0 || i>= num_samples())
    {
      cerr << "Attempt to access sample " << i << " of a " << num_samples() << " sample wave.\n";
      if (num_samples()>0)
	return *(p_values.error_return);
    }

  if (channel<0 || channel>= num_channels())
    {
      cerr << "Attempt to access channel " << channel << " of a " << num_channels() << " channel wave.\n";
      if (num_samples()>0)
	return *(p_values.error_return);
    }

  return p_values.a_no_check(i,channel);
}

short EST_Wave::a(int i, int channel) const
{
  return ((EST_Wave *)this)->a(i,channel);
}

short &EST_Wave::a_safe(int i, int channel)
{
    static short out_of_bound_value = 0;

    if ((i < 0) || (i >= num_samples()))
    {   // need to give them something but they might have changed it
	// so reinitialise it to 0 first
	out_of_bound_value = 0;
	return out_of_bound_value;
    }
    else
	return a_no_check(i,channel);
}

void EST_Wave::fill(short v, int channel)
{
    if (channel == EST_ALL)
      p_values.fill(v);
    else 
      for (int i = 0; i < num_samples(); ++i)
	p_values.a_no_check(i,channel) = v;
}

EST_read_status EST_Wave::load(const EST_String filename, 
			       int offset, int length,
			       int rate)
{
    EST_read_status stat = read_error;
    EST_TokenStream ts;

    if ((ts.open(filename)) == -1)
    {
	cerr << "Wave load: can't open file \"" << filename << "\"" << endl;
	return stat;
    }

    stat = load(ts,offset,length,rate);
    ts.close();
    return stat;
}

EST_read_status EST_Wave::load(EST_TokenStream &ts,
			       int offset, int length,
			       int rate)
{
    EST_read_status stat = read_error;
    int pos = ts.tell();
    
    for(int n=0; n< EST_WaveFile::map.n() ; n++)
    {
	EST_WaveFileType t = EST_WaveFile::map.token(n);
	
	if (t == wff_none)
	    continue;
	
	EST_WaveFile::Info *info = &(EST_WaveFile::map.info(t));
	
	if (! info->recognise)
	    continue;
	
	EST_WaveFile::Load_TokenStream * l_fun =info->load;
	
	if (l_fun == NULL)
	    continue;

	ts.seek(pos);
	stat = (*l_fun)(ts, *this, 
			rate, st_short, EST_NATIVE_BO, 1,
			offset, length);
	
	if (stat == read_ok)
	{
	    set_file_type(EST_WaveFile::map.value(t));
	    break;
	}
    }
    
    return stat;
}

EST_read_status EST_Wave::load(const EST_String filename, 
			       const EST_String type, 
			       int offset, int length,
			       int rate)
{
    EST_read_status stat = read_error;
    EST_TokenStream ts;

    if (filename == "-")
	ts.open(stdin,FALSE);
    else if ((ts.open(filename)) == -1)
    {
	cerr << "Wave load: can't open file \"" << filename << "\"" << endl;
	return stat;
    }

    stat = load(ts,type,offset,length,rate);
    ts.close();
    return stat;
}

EST_read_status EST_Wave::load(EST_TokenStream &ts,
			       const EST_String type, 
			       int offset, int length,
			       int rate)
{
    EST_WaveFileType t = EST_WaveFile::map.token(type);

    if (t == wff_none)
    {
	cerr << "Unknown Wave file type " << type << endl;
	return read_error;
    }

    EST_WaveFile::Load_TokenStream * l_fun = EST_WaveFile::map.info(t).load;
    
    if (l_fun == NULL)
    {
	cerr << "Can't load waves to files type " << type << endl;
	return read_error;
    }
    
    set_file_type(EST_WaveFile::map.value(t));
    return (*l_fun)(ts, *this,
		    rate, st_short, EST_NATIVE_BO, 1,
		    offset, length);
   
}

EST_read_status EST_Wave::load_file(const EST_String filename, 
		     const EST_String type, int sample_rate,
		     const EST_String stype, int bov, int nc, int offset,
		     int length)
{
    EST_read_status stat = read_error;
    EST_TokenStream ts;

    if (filename == "-")
	ts.open(stdin,FALSE);
    else if ((ts.open(filename)) == -1)
    {
	cerr << "Wave load: can't open file \"" << filename << "\"" << endl;
	return stat;
    }

    stat = load_file(ts,type,sample_rate,stype,bov,nc,offset,length);
    ts.close();
    return stat;
}

EST_read_status EST_Wave::load_file(EST_TokenStream &ts,
		     const EST_String type, int sample_rate,
		     const EST_String stype, int bov, int nc, int offset,
		     int length)

{
  EST_WaveFileType t = EST_WaveFile::map.token(type);

  EST_sample_type_t values_type = sample_type_map.token(stype);

  if (t == wff_none)
    {
      cerr << "Unknown Wave file type " << type << endl;
      return read_error;
    }

  EST_WaveFile::Load_TokenStream * l_fun = EST_WaveFile::map.info(t).load;
    
  if (l_fun == NULL)
    {
      cerr << "Can't load waves to files type " << type << endl;
      return read_error;
    }
    
  return (*l_fun)(ts, *this,
		  sample_rate, values_type, bov, nc,
		  offset, length);
    
}    

void EST_Wave::sub_wave(EST_Wave &sw, 
			int offset, int num,
			int start_c, int nchan)
{
  if (num == EST_ALL)
    num = num_samples()-offset;
  if (nchan == EST_ALL)
    nchan = num_channels()-start_c;

  p_values.sub_matrix(sw.p_values, offset, num, start_c, nchan);
  sw.set_sample_rate(sample_rate());
}

EST_write_status EST_Wave::save(const EST_String filename, 
				 const EST_String type)
{
    FILE *fp;

    if (filename == "-")
	fp = stdout;
    else if ((fp = fopen(filename,"wb")) == NULL)
    {
	cerr << "Wave save: can't open output file \"" <<
	    filename << "\"" << endl;
	return write_fail;
    }

    EST_write_status r = save(fp,type);
    if (fp != stdout)
	fclose(fp);
    return r;
}

EST_write_status EST_Wave::save(FILE *fp, const EST_String type)
{
    EST_String save_type = (type == "") ? DEF_FILE_TYPE : type;

    EST_WaveFileType t = EST_WaveFile::map.token(save_type);

    if (t == wff_none)
    {
	cerr << "Wave: unknown filetype in saving " << save_type << endl;
	return write_fail;
    }
    
    EST_WaveFile::Save_TokenStream * s_fun = EST_WaveFile::map.info(t).save;
    
    if (s_fun == NULL)
    {
	cerr << "Can't save waves to files type " << save_type << endl;
	return write_fail;
    }
    
    return (*s_fun)(fp, *this, st_short, EST_NATIVE_BO);
}

EST_write_status EST_Wave::save_file(const EST_String filename,
				     EST_String ftype,
				     EST_String stype, int obo)
{
    FILE *fp;

    if (filename == "-")
	fp = stdout;
    else if ((fp = fopen(filename,"wb")) == NULL)
    {
	cerr << "Wave save: can't open output file \"" <<
	    filename << "\"" << endl;
	return write_fail;
    }

    EST_write_status r = save_file(fp,ftype,stype,obo);
    if (fp != stdout)
	fclose(fp);
    return r;
}

EST_write_status EST_Wave::save_file(FILE *fp,
				     EST_String ftype,
				     EST_String stype, int obo)
{
    EST_WaveFileType t = EST_WaveFile::map.token(ftype);
    EST_sample_type_t sample_type = sample_type_map.token(stype);

    if (t == wff_none)
      {
	cerr << "Unknown Wave file type " << ftype << endl;
	return write_fail;
      }

    EST_WaveFile::Save_TokenStream * s_fun = EST_WaveFile::map.info(t).save;
    
    if (s_fun == NULL)
    {
	cerr << "Can't save waves to files type " << ftype << endl;
	return write_fail;
    }
    
    return (*s_fun)(fp, *this, sample_type, obo);
    
}

void EST_Wave::resample(int new_freq)
{
    // Rseample wave to new sample rate
  if (new_freq != p_sample_rate)
    {
      if (p_values.rateconv(p_sample_rate, new_freq) != 0)
	{
	  cerr << "rateconv: failed to convert\n";
	  return;
	}
	set_sample_rate(new_freq);
    }
    
}

void EST_Wave::rescale(float gain, int normalize)
{
    int ns;
    float factor = gain;
    
    if (normalize)
    {
	int max = 0;
	for (int i = 0; i < num_samples(); ++i)
	  for (int j = 0; j < num_channels(); ++j)
	    if (abs(a_no_check(i,j)) > max)
		max = abs(a_no_check(i,j));
	factor *= 32766.0/(float)max;
    }
    
    for (int i = 0; i < num_samples(); ++i)
      for (int j = 0; j < num_channels(); ++j)
	{
	  ns = (int) ((float)a_no_check(i,j) * factor + 0.5);
	  if (ns < -32766)
	    a_no_check(i,j)= -32766;
	  else if (ns > 32766)
	    a_no_check(i,j)= 32766;
	  else
	    a_no_check(i,j)= ns;
	}
}

EST_Wave &EST_Wave::operator =(const EST_Wave &w)
{
    copy(w);
    return *this;
}

EST_Wave &EST_Wave::operator +=(const EST_Wave &w)
{
    EST_Wave w2;
    const EST_Wave *toadd = &w;

    if (w.num_channels() != num_channels())
    {
	cerr << "Cannot concatenate waveforms with differing numbers of channels\n";
	return *this;
    }

    if (p_sample_rate != w.sample_rate())
    {
	w2 = w;
	w2.resample(p_sample_rate);
	toadd= &w2;
    }

    p_values.add_rows(toadd->p_values);

    return *this;
}

// add wave p_values to existing wave in parallel to create multi-channel wave.
EST_Wave &EST_Wave::operator |=(const EST_Wave &wi)
{
    int i, k;
    EST_Wave w = wi; // to allow resampling of const

    w.resample(p_sample_rate);  // far too difficult otherwise

    int o_channels = num_channels();
    int r_channels = num_channels()+w.num_channels();
    int r_samples = Gof(num_samples(), w.num_samples());

    resize(r_samples, r_channels);

    for (k = 0; k < w.num_channels(); ++k)
	for (i=0; i < w.num_samples(); i++)
	    a(i,k+o_channels) += w.a(i, k);

    return *this;
}

ostream& operator << (ostream& p_values, const EST_Wave &sig)
{
    for (int i = 0; i < sig.num_samples(); ++i)
	p_values << sig(i) << "\n";
    
    return p_values;
}
